"""Messages for wake word detection."""
import re
import typing
from dataclasses import dataclass
from enum import Enum

from dataclasses_json import LetterCase, dataclass_json

from .base import Message


class HotwordToggleReason(str, Enum):
    """Reason for hotword toggle on/off."""

    UNKNOWN = ""
    """Overrides all other reasons."""
    DIALOGUE_SESSION = "dialogueSession"
    """Dialogue session is active."""
    PLAY_AUDIO = "playAudio"
    """Audio is currently playing."""
    TTS_SAY = "ttsSay"
    """Text to speech system is currently speaking."""


@dataclass
class HotwordToggleOn(Message):
    """Activate the wake word component, so pronouncing a wake word will trigger a
    :class:`HotwordDetected` message.

    .. admonition:: MQTT message

      Topic
        ``hermes/hotword/toggleOn``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - siteId
            - String
            - The id of the site where the wake word component should be enabled.
          * - reason
            - String
            - The reason for enabling the wake word component.

      Publish this message type with ``mosquitto_pub``:

      .. code-block:: shell

        mosquitto_pub -h <HOSTNAME> -t 'hermes/hotword/toggleOn' -m '{"siteId": "default", "reason": "dialogueSession"}'
    """

    site_id: str = "default"
    """The id of the site where the wake word component should be enabled."""

    # ------------
    # Rhasspy only
    # ------------

    reason: HotwordToggleReason = HotwordToggleReason.UNKNOWN
    """The reason for enabling the wake word component.

    Note
    ----

    This is a Rhasspy-only attribute.
    """

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"hermes/hotword/toggleOn"``
        """
        return "hermes/hotword/toggleOn"


@dataclass
class HotwordToggleOff(Message):
    """Deactivate the wake word component, so pronouncing a wake word won't trigger a
    :class:`HotwordDetected` message.

    .. admonition:: MQTT message

      Topic
        ``hermes/hotword/toggleOff``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - siteId
            - String
            - The id of the site where the wake word component should be disabled.
          * - reason
            - String
            - The reason for disabling the wake word component.

      Publish this message type with ``mosquitto_pub``:

      .. code-block:: shell

        mosquitto_pub -h <HOSTNAME> -t 'hermes/hotword/toggleOff' -m '{"siteId": "default", "reason": "dialogueSession"}'
    """

    site_id: str = "default"
    """The id of the site where the wake word component should be disabled."""

    # ------------
    # Rhasspy only
    # ------------

    reason: HotwordToggleReason = HotwordToggleReason.UNKNOWN
    """The reason for disabling the wake word component.

    Note
    ----

    This is a Rhasspy-only attribute.
    """

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"hermes/hotword/toggleOff"``
        """
        return "hermes/hotword/toggleOff"


@dataclass
class HotwordDetected(Message):
    """Message sent by the wake word component when it has detected a specific wake word.

    .. admonition:: MQTT message

      Topic
        ``hermes/hotword/<WAKEWORD_ID>/detected``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - modelId
            - String
            - The id of the model that triggered the wake word.
          * - modelVersion
            - String
            - The version of the model.
          * - modelType
            - String
            - The type of the model. Possible values are ``"universal"`` and ``"personal"``.
          * - currentSensitivity
            - Float
            - The sensitivity configured in the model at the time of the detection.
          * - siteId
            - String
            - The id of the site where the wake word component should be disabled.
          * - sessionId
            - String (optional)
            - The id of the dialogue session created after detection.
          * - send_audio_captured
            - Boolean (optional)
            - ``True`` if audio captured from the ASR should be emitted on
              ``rhasspy/asr/{site_id}/{session_id}/audioCaptured``.
          * - lang
            - String (optional)
            - Language of the detected wake word.
              Copied by the dialogue manager into subsequent ASR and NLU messages.

      Subscribe to this message type with ``mosquitto_sub``:

      .. code-block:: shell

        mosquitto_sub -h <HOSTNAME> -v -t 'hermes/hotword/default/detected'
    """

    TOPIC_PATTERN = re.compile(r"^hermes/hotword/([^/]+)/detected$")

    model_id: str
    """The id of the model that triggered the wake word."""
    model_version: str = ""
    """The version of the model."""
    model_type: str = "personal"
    """The type of the model. Possible values are ``"universal"`` and ``"personal"``."""
    current_sensitivity: float = 1.0
    """The sensitivity configured in the model at the time of the detection."""
    site_id: str = "default"
    """The id of the site where the wake word was detected."""

    # ------------
    # Rhasspy only
    # ------------

    session_id: typing.Optional[str] = None
    """The desired id of the dialogue session created after detection.
    Leave empty to have one auto-generated.

    Note
    ----

    This is a Rhasspy-only attribute.
    """
    send_audio_captured: typing.Optional[bool] = None
    """``True`` if audio captured from the ASR should be emitted on
    ``rhasspy/asr/{site_id}/{session_id}/audioCaptured``.

    Note
    ----

    This is a Rhasspy-only attribute.
    """
    lang: typing.Optional[str] = None
    """Language of the detected wake word.
    Copied by the dialogue manager into subsequent ASR and NLU messages.

    Note
    ----

    This is a Rhasspy-only attribute.
    """
    custom_entities: typing.Optional[typing.Dict[str, typing.Any]] = None
    """User-defined entities to be set in the recognized intent.
    Copied by the dialogue manager into subsequent ASR and NLU messages.

    Note
    ----

    This is a Rhasspy-only attribute.
    """

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Arguments
        ---------
        wakeword_id: str
            The id of the wake word.

        Returns
        -------
        str
            MQTT topic for this message type with the given wake word id.

        Example
        -------
        >>> from rhasspyhermes.wake import HotwordDetected
        >>> HotwordDetected.topic(wakeword_id="example-02.wav")
        'hermes/hotword/example-02.wav/detected'
        """
        wakeword_id = kwargs.get("wakeword_id", "+")
        return f"hermes/hotword/{wakeword_id}/detected"

    @classmethod
    def get_wakeword_id(cls, topic: str) -> str:
        """Get wakeword id from MQTT topic.

        Arguments
        ---------
        topic
            MQTT topic.

        Returns
        -------
        str
            Wake word ID extracted from the MQTT topic.

        Example
        -------
        >>> from rhasspyhermes.wake import HotwordDetected
        >>> HotwordDetected.get_wakeword_id("hermes/hotword/example-02.wav/detected")
        'example-02.wav'
        """
        match = re.match(HotwordDetected.TOPIC_PATTERN, topic)
        assert match, "Not a detected topic"
        return match.group(1)

    @classmethod
    def is_topic(cls, topic: str) -> bool:
        """True if topic matches template."""
        return re.match(HotwordDetected.TOPIC_PATTERN, topic) is not None


# -----------------------------------------------------------------------------
# Rhasspy Only Messages
# -----------------------------------------------------------------------------


@dataclass
class HotwordError(Message):
    """Error from wake word component.

    .. admonition:: MQTT message

      Topic
        ``hermes/error/hotword``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - error
            - String
            - A description of the error that occurred.
          * - siteId
            - String
            - The id of the site where the error occurred.
          * - context
            - String (optional)
            - Additional information on the context in which the error occurred.
          * - sessionId
            - String (optional)
            - The id of the session, if there is an active session.

      Subscribe to this message type with ``mosquitto_sub``:

      .. code-block:: shell

        mosquitto_sub -h <HOSTNAME> -v -t 'hermes/error/hotword'

    Note
    ----

    This is a Rhasspy-only message.
    """

    error: str
    """A description of the error that occurred."""
    site_id: str = "default"
    """The id of the site where the error occurred."""
    context: typing.Optional[str] = None
    """Additional information on the context in which the error occurred."""
    session_id: typing.Optional[str] = None
    """The id of the session, if there is an active session."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"hermes/error/hotword"``
        """
        return "hermes/error/hotword"


@dataclass
class GetHotwords(Message):
    """Request to list available hotwords. The wake word component responds with a
    :class:`Hotwords` message.

    .. admonition:: MQTT message

      Topic
        ``rhasspy/hotword/getHotwords``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - siteId
            - String
            - The id of the site where the wake word component exists.
          * - id
            - String (optional)
            - Unique id passed to the response in the :class:`Hotwords` message.

      Publish this message type with ``mosquitto_pub``:

      .. code-block:: shell

        mosquitto_pub -h <HOSTNAME> -t 'rhasspy/hotword/getHotwords' -m '{"siteId": "default", "id": "foobar"}'

    Note
    ----

    This is a Rhasspy-only message.
    """

    site_id: str = "default"
    """The id of the site where the wake word component exists."""
    id: typing.Optional[str] = None
    """Unique id passed to the response in the :class:`Hotwords` message."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"rhasspy/hotword/getHotwords"``
        """
        return "rhasspy/hotword/getHotwords"


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Hotword:
    """Description of a single hotword."""

    model_id: str
    """Unique ID of hotword model."""
    model_words: str
    """Actual words used to activate hotword."""
    model_version: str = ""
    """Model version."""
    model_type: str = "personal"
    """Model type (personal, unversal)."""


@dataclass
class Hotwords(Message):
    """The list of available hotwords. The wake word component sends this message
    in response to a request in a :class:`GetHotwords` message.

    .. admonition:: MQTT message

      Topic
        ``rhasspy/hotword/hotwords``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - models
            - List of JSON objects
            - The list of available hotwords.
          * - siteId
            - String
            - The id of the site where hotwords were requested.
          * - id
            - String (optional)
            - Unique id passed from the request in the :class:`GetHotwords` message.

      Subscribe to this message type with ``mosquitto_sub``:

      .. code-block:: shell

        mosquitto_sub -h <HOSTNAME> -v -t 'rhasspy/hotword/hotwords'

    Note
    ----

    This is a Rhasspy-only message.
    """

    models: typing.List[Hotword]
    """The list of available hotwords."""
    site_id: str = "default"
    """The id of the site where hotwords were requested."""
    id: typing.Optional[str] = None
    """Unique id passed from the request in the :class:`GetHotwords` message."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"rhasspy/hotword/hotwords"``
        """
        return "rhasspy/hotword/hotwords"


@dataclass
class RecordHotwordExample(Message):
    """Request to record examples of a hotword. The wake word component responds with a
    :class:`HotwordExampleRecorded` message.

    .. admonition:: MQTT message

      Topic
        ``rhasspy/hotword/recordExample``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - id
            - String
            - Unique id used in the response message (:class:`HotwordExampleRecorded`).
          * - siteId
            - String
            - The id of the site where the wake word component exists.

      Publish this message type with ``mosquitto_pub``:

      .. code-block:: shell

        mosquitto_pub -h <HOSTNAME> -t 'rhasspy/hotword/recordExample' -m '{"siteId": "default", "id": "foobar"}'

    Note
    ----

    This is a Rhasspy-only message.
    """

    id: str
    """Unique id used in the response message (:class:`HotwordExampleRecorded`)."""
    site_id: str = "default"
    """The id of the site where the wake word component exists."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"rhasspy/hotword/recordExample"``
        """
        return "rhasspy/hotword/recordExample"


@dataclass
class HotwordExampleRecorded(Message):
    """Response when a hotword example has been recorded. Sent by the wake word component
    in response to a :class:`RecordHotwordExample` message.

    .. admonition:: MQTT message

      Topic
        ``rhasspy/hotword/<SITE_ID>/exampleRecorded/<REQUEST_ID>``

      Payload (binary)
        Audio from the recorded sample in WAV format.

      Subscribe to this message type with ``mosquitto_sub`` and show the binary payload as hexadecimal numbers:

      .. code-block:: shell

        mosquitto_sub -h <HOSTNAME> -t 'rhasspy/hotword/<SITE_ID>/exampleRecorded/<REQUEST_ID>' -F %x

    Note
    ----

    This is a Rhasspy-only message.
    """

    TOPIC_PATTERN = re.compile(r"^rhasspy/hotword/([^/]+)/exampleRecorded/([^/]+)$")

    wav_bytes: bytes
    """Audio from recorded sample in WAV format."""

    def payload(self) -> typing.Union[str, bytes]:
        """Get binary/string for this message."""
        return self.wav_bytes

    @classmethod
    def is_binary_payload(cls) -> bool:
        """True if payload is not JSON.

        Returns
        -------
        bool
            ``True``
        """
        return True

    @classmethod
    def is_site_in_topic(cls) -> bool:
        """True if site id is in topic.

        Returns
        -------
        bool
            ``True``
        """
        return True

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Arguments
        ---------
        site_id: str
            The id of the site where the wake word component exists.
        request_id: str
            Unique id of the request message.

        Returns
        -------
        str
            MQTT topic for this message type with the given site id and request id.

        Example
        -------
        >>> from rhasspyhermes.wake import HotwordExampleRecorded
        >>> HotwordExampleRecorded.topic(site_id="default", request_id="foobar")
        'rhasspy/hotword/default/exampleRecorded/foobar'
        """
        site_id = kwargs.get("site_id", "+")
        request_id = kwargs.get("request_id", "#")
        return f"rhasspy/hotword/{site_id}/exampleRecorded/{request_id}"

    @classmethod
    def get_site_id(cls, topic: str) -> typing.Optional[str]:
        """Get site id from MQTT topic.

        Arguments
        ---------
        topic
            MQTT topic.

        Returns
        -------
        str
            Site ID extracted from the MQTT topic.

        Example
        -------
        >>> from rhasspyhermes.wake import HotwordExampleRecorded
        >>> HotwordExampleRecorded.get_site_id("rhasspy/hotword/default/exampleRecorded/foobar")
        'default'
        """
        match = re.match(HotwordExampleRecorded.TOPIC_PATTERN, topic)
        assert match, "Not an exampleRecorded topic"
        return match.group(1)

    @classmethod
    def get_request_id(cls, topic: str) -> str:
        """Get request id from MQTT topic.

        Arguments
        ---------
        topic
            MQTT topic.

        Returns
        -------
        str
            Request ID extracted from the MQTT topic.

        Example
        -------
        >>> from rhasspyhermes.wake import HotwordExampleRecorded
        >>> HotwordExampleRecorded.get_request_id("rhasspy/hotword/default/exampleRecorded/foobar")
        'foobar'
        """
        match = re.match(HotwordExampleRecorded.TOPIC_PATTERN, topic)
        assert match, "Not an exampleRecorded topic"
        return match.group(2)

    @classmethod
    def is_topic(cls, topic: str) -> bool:
        """True if topic matches template"""
        return re.match(HotwordExampleRecorded.TOPIC_PATTERN, topic) is not None



"""Messages for looking up/guessing word pronunciations."""
import typing
from dataclasses import dataclass

from dataclasses_json import LetterCase, dataclass_json

from rhasspyhermes.base import Message


@dataclass
class G2pPronounce(Message):
    """Get phonetic pronunciation for words.

    The response is sent in a :class:`G2pPhonemes` message.

    .. admonition:: MQTT message

      Topic
        ``rhasspy/g2p/pronounce``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - words
            - List of Strings
            - Words to guess pronunciations for.
          * - id
            - String (optional)
            - Unique id for request. Appended to reply topic (:class:`G2pPhonemes`).
          * - siteId
            - String
            - The id of the site where pronunciations were requested. Defaults to ``"default"``.
          * - sessionId
            - String (optional)
            - Id of active session, if there is one.
          * - numGuesses
            - Integer
            - Maximum number of guesses to return for words not in dictionary. Defaults to 5.

      Publish this message type with ``mosquitto_pub``:

      .. code-block:: shell

        mosquitto_pub -h <HOSTNAME> -t 'rhasspy/g2p/pronounce' -m '{"words": ["word", "sentence"], "id": "test", "siteId": "default", "sessionId": null, "numGuesses": 5}'

    Example
    -------

    >>> from rhasspyhermes.g2p import G2pPronounce
    >>> p = G2pPronounce(words=["word", "sentence"], id="test")
    >>> p.payload()
    '{"words": ["word", "sentence"], "id": "test", "siteId": "default", "sessionId": null, "numGuesses": 5}'
    >>> p.topic()
    'rhasspy/g2p/pronounce'

    Note
    ----

    This is a Rhasspy-only message."""

    words: typing.List[str]
    """Words to guess pronunciations for."""
    id: typing.Optional[str] = None
    """Unique id for request. Appended to reply topic (:class:`G2pPhonemes`)."""
    site_id: str = "default"
    """Id of site to request pronunciations from."""
    session_id: typing.Optional[str] = None
    """Id of active session, if there is one."""
    num_guesses: int = 5
    """Maximum number of guesses to return for words not in dictionary."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"rhasspy/g2p/pronounce"``
        """
        return "rhasspy/g2p/pronounce"


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class G2pPronunciation:
    """Phonetic pronunciation for a single word."""

    phonemes: typing.List[str]
    """Phonetic pronunciation for word."""
    guessed: typing.Optional[bool] = None
    """``True`` if this pronunciation was guessed using a g2p model.
    ``False`` if it came from a pronunciation dictionary."""


@dataclass
class G2pPhonemes(Message):
    """Response to :class:`G2pPronounce`.

    .. admonition:: MQTT message

      Topic
        ``rhasspy/g2p/phonemes``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - wordPhonemes
            - Dictionary
            - Guessed or looked up pronunciations.
          * - id
            - String (optional)
            - Unique id for a :class:`G2pPronounce` request.
          * - siteId
            - String
            - The id of the site where pronunciations were requested. Defaults to ``"default"``.
          * - sessionId
            - String (optional)
            - Id of active session, if there is one.

      Subscribe to this message type with ``mosquitto_sub``:

      .. code-block:: shell

        mosquitto_sub -h <HOSTNAME> -t 'rhasspy/g2p/phonemes' -v

    Note
    ----

    This is a Rhasspy-only message."""

    word_phonemes: typing.Dict[str, typing.List[G2pPronunciation]]
    """Guessed or looked up pronunciations."""
    id: typing.Optional[str] = None
    """Unique id from a :class:`G2pPronounce` request."""
    site_id: str = "default"
    """Id of site where pronunciations were requested."""
    session_id: typing.Optional[str] = None
    """Id of active session, if there is one."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"rhasspy/g2p/phonemes"``

        """
        return "rhasspy/g2p/phonemes"


@dataclass
class G2pError(Message):
    """Error from G2P component.

    .. admonition:: MQTT message

      Topic
        ``rhasspy/error/g2p``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - error
            - String
            - A description of the error that occurred.
          * - siteId
            - String
            - The id of the site where the error occurred.
          * - context
            - String
            - Additional information on the context in which the error occurred.
          * - sessionId
            - String (optional)
            - Id of active session, if there is one.

      Subscribe to this message type with ``mosquitto_sub``:

      .. code-block:: shell

        mosquitto_sub -h <HOSTNAME> -t 'rhasspy/error/g2p' -v

    Note
    ----

    This is a Rhasspy-only message.
    """

    error: str
    """A description of the error that occurred."""
    site_id: str = "default"
    """The id of the site where the error occurred."""
    context: typing.Optional[str] = None
    """Additional information on the context in which the error occurred."""
    session_id: typing.Optional[str] = None
    """The id of the session, if there is an active session."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"rhasspy/error/g2p"``
        """
        return "rhasspy/error/g2p"



"""Messages for the Hermes dialogue manager."""
import typing
from collections.abc import Mapping
from dataclasses import dataclass
from enum import Enum

from dataclasses_json import DataClassJsonMixin, LetterCase, dataclass_json
from dataclasses_json.core import Json

from .base import Message


class DialogueActionType(str, Enum):
    """Type of session init objects."""

    ACTION = "action"
    """Use this type when you need the user to respond."""
    NOTIFICATION = "notification"
    """Use this type when you only want to inform the user of something without
    expecting a response."""


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class DialogueAction(DataClassJsonMixin):
    """Dialogue session action."""

    can_be_enqueued: bool
    """If true, the session will start when there is no pending one on this
    site. Otherwise, the session is just dropped if there is running one."""
    type: DialogueActionType = DialogueActionType.ACTION
    """This value is always :class:`DialogueActionType.ACTION`."""
    text: typing.Optional[str] = None
    """Text that the TTS should say at the beginning of the session."""
    intent_filter: typing.Optional[typing.List[str]] = None
    """A list of intents names to restrict the NLU resolution on the first query."""
    send_intent_not_recognized: bool = False
    """Indicates whether the dialogue manager should handle non-recognized
    intents by itself or send them for the client to handle."""


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class DialogueNotification(DataClassJsonMixin):
    """Dialogue session notification."""

    text: str
    """Text the TTS should say."""
    type: DialogueActionType = DialogueActionType.NOTIFICATION
    """This value is always :class:`DialogueActionType.NOTIFICATION`."""


class DialogueSessionTerminationReason(str, Enum):
    """The reason why the session was ended."""

    NOMINAL = "nominal"
    """The session ended as expected (a :class:`DialogueEndSession` message was received)."""
    ABORTED_BY_USER = "abortedByUser"
    """The session was aborted by the user."""
    INTENT_NOT_RECOGNIZED = "intentNotRecognized"
    """The session ended because no intent was successfully detected."""
    TIMEOUT = "timeout"
    """The session timed out because there was no response from one of the components or
    no :class:`DialogueContinueSession` or :class:`DialogueEndSession` message in a timely manner."""
    ERROR = "error"
    """The session failed with an error."""


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class DialogueSessionTermination:
    """Dialogue session termination type."""

    reason: DialogueSessionTerminationReason
    """The reason why the session was ended."""


# -----------------------------------------------------------------------------


@dataclass
class DialogueStartSession(Message):
    """Start a dialogue session.

    You can send this message to programmatically initiate a new session. The
    Dialogue Manager will start the session by asking the TTS to say the text
    (if any) and wait for the answer of the user.

    .. admonition:: MQTT message

      Topic
        ``hermes/dialogueManager/startSession``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - siteId
            - String
            - The id of the site where to start the session.
          * - init
            - JSON object
            - Session initialization description.
          * - customData
            - String (optional)
            - Additional information that can be provided by the handler. Each message
              related to the new session - sent by the Dialogue Manager - will contain
              this data.
          * - lang
            - String (optional)
            - Language of the session.

      Publish this message type with ``mosquitto_pub``:

      .. code-block:: shell

        mosquitto_pub -h <HOSTNAME> -t 'hermes/dialogueManager/startSession' -m '{"siteId": "livingroom", "init": {"type": "notification", "text": "Ready"}, "lang": "en"}'

    Example
    -------
    >>> from rhasspyhermes.dialogue import DialogueStartSession, DialogueNotification
    >>> start_session = DialogueStartSession(init=DialogueNotification(text="Ready"), site_id="livingroom", lang="en")
    >>> start_session
    DialogueStartSession(init=DialogueNotification(text='Ready', type=<DialogueActionType.NOTIFICATION: 'notification'>), site_id='livingroom', custom_data=None, lang='en')
    >>> start_session.payload()
    '{"init": {"text": "Ready", "type": "notification"}, "siteId": "livingroom", "customData": null, "lang": "en"}'
    """

    init: typing.Union[DialogueAction, DialogueNotification]
    """Session initialization description."""
    site_id: str = "default"
    """The id of the site where to start the session."""
    custom_data: typing.Optional[str] = None
    """Additional information that can be provided by the handler. Each message
    related to the new session - sent by the Dialogue Manager - will contain
    this data."""
    lang: typing.Optional[str] = None
    """Language of the session.

    Note
    ----

    This is a Rhasspy-only attribute.
    """

    # pylint: disable=W0221,W0237
    @classmethod
    def from_dict(
        cls: typing.Type["DialogueStartSession"],
        message_dict: Json,
        *,
        infer_missing=False
    ) -> "DialogueStartSession":
        assert isinstance(message_dict, Mapping)
        init = message_dict.pop("init")
        if init["type"] == DialogueActionType.NOTIFICATION:
            message_dict["init"] = DialogueNotification.from_dict(init)
        else:
            message_dict["init"] = DialogueAction.from_dict(init)

        return super(DialogueStartSession, cls).from_dict(
            message_dict, infer_missing=infer_missing
        )

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"hermes/dialogueManager/startSession"``
        """
        return "hermes/dialogueManager/startSession"


@dataclass
class DialogueSessionQueued(Message):
    """Sent by the dialogue manager when it receives a :class:`DialogueStartSession` message
    and the site where the interaction should take place is busy. When the site is free again,
    the session will be started.

    Only :class:`DialogueStartSession` messages with an ``init`` attribute of the following type
    can be enqueued:

    - :class:`DialogueNotification`
    - :class:`DialogueAction` with the attribute ``canBeEnqueued`` set to ``True``.

    .. admonition:: MQTT message

      Topic
        ``hermes/dialogueManager/sessionQueued``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - sessionId
            - String
            - The id of the session that was enqueued.
          * - siteId
            - String
            - The id of the site where the user interaction will take place.
          * - customData
            - String (optional)
            - Custom data provided in the :class:`DialogueStartSession` message.

      Subscribe to this message type with ``mosquitto_sub``:

      .. code-block:: shell

        mosquitto_sub -h <HOSTNAME> -v -t 'hermes/dialogueManager/sessionQueued'
    """

    session_id: str
    """The id of the session that was enqueued."""
    site_id: str = "default"
    """The id of the site where the user interaction will take place."""
    custom_data: typing.Optional[str] = None
    """Custom data provided in the :class:`DialogueStartSession` message."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"hermes/dialogueManager/sessionQueued"``
        """
        return "hermes/dialogueManager/sessionQueued"


@dataclass
class DialogueSessionStarted(Message):
    """Sent when a dialogue session has been started.

    .. admonition:: MQTT message

      Topic
        ``hermes/dialogueManager/sessionStarted``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - sessionId
            - String
            - The id of the session that was started.
          * - siteId
            - String
            - The id of the site where the user interaction is taking place.
          * - customData
            - String (optional)
            - Custom data provided in the :class:`DialogueStartSession` message.
          * - lang
            - String (optional)
            - Language of the session.

      Subscribe to this message type with ``mosquitto_sub``:

      .. code-block:: shell

        mosquitto_sub -h <HOSTNAME> -v -t 'hermes/dialogueManager/sessionStarted'
    """

    session_id: str
    """The id of the session that was started."""
    site_id: str = "default"
    """The id of the site where the user interaction is taking place."""
    custom_data: typing.Optional[str] = None
    """Custom data provided in the :class:`DialogueStartSession` message."""
    lang: typing.Optional[str] = None
    """Language of the session.

    Note
    ----

    This is a Rhasspy-only attribute.
    """

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"hermes/dialogueManager/sessionStarted"``

        """
        return "hermes/dialogueManager/sessionStarted"


@dataclass
class DialogueContinueSession(Message):
    """Sent when a dialogue session should be continued.

    You should send this message after receiving a :class:`rhasspyhermes.nlu.NluIntent` message
    if you want to continue the session. This can be used for example to ask
    additional information to the user.

    Make sure to use the same ``sessionId`` as the original :class:`rhasspyhermes.nlu.NluIntent`
    message.

    .. admonition:: MQTT message

      Topic
        ``hermes/dialogueManager/continueSession``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - sessionId
            - String
            - The id of the session to continue.
          * - customData
            - String (optional)
            - An update to the session's custom data. If not provided, the custom data
              will stay the same.
          * - text
            - String (optional)
            - The text the TTS should say to start this additional request of the
              session.
          * - intentFilter
            - List of Strings (optional)
            - A list of intent names to restrict the NLU resolution on the answer of
              this query.
          * - sendIntentNotRecognized
            - Boolean
            - Indicates whether the dialogue manager should handle non recognized
              intents by itself or send them for the client to handle.
          * - slot
            - String (optional)
            - Unused.
          * - lang
            - String (optional)
            - Language of the session.
              Leave empty to use setting from start of session.

      Publish this message type with ``mosquitto_pub``:

      .. code-block:: shell

        mosquitto_pub -h <HOSTNAME> -t 'hermes/dialogueManager/continueSession' -m '{"sessionId": "foobar", "customData": null, "text": "Are you sure?", "intentFilter": null, "sendIntentNotRecognized": false, "slot": null, "lang": null}'

    Example
    -------
    >>> from rhasspyhermes.dialogue import DialogueContinueSession
    >>> session = DialogueContinueSession(session_id="foobar", text="Are you sure?")
    >>> session
    DialogueContinueSession(session_id='foobar', custom_data=None, text='Are you sure?', intent_filter=None, send_intent_not_recognized=False, slot=None, lang=None)
    >>> session.payload()
    '{"sessionId": "foobar", "customData": null, "text": "Are you sure?", "intentFilter": null, "sendIntentNotRecognized": false, "slot": null, "lang": null}'
    """

    session_id: str
    """The id of the session to continue."""
    custom_data: typing.Optional[str] = None
    """An update to the session's custom data. If not provided, the custom data
    will stay the same."""
    text: typing.Optional[str] = None
    """The text the TTS should say to start this additional request of the
    session."""
    intent_filter: typing.Optional[typing.List[str]] = None
    """A list of intent names to restrict the NLU resolution on the answer of
    this query."""
    send_intent_not_recognized: bool = False
    """Indicates whether the dialogue manager should handle non recognized
    intents by itself or send them for the client to handle."""
    slot: typing.Optional[str] = None
    """Unused."""

    # ------------
    # Rhasspy only
    # ------------

    lang: typing.Optional[str] = None
    """Language of the session.
    Leave empty to use setting from start of session.

    Note
    ----

    This is a Rhasspy-only attribute.
    """

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"hermes/dialogueManager/continueSession"``
        """
        return "hermes/dialogueManager/continueSession"


@dataclass
class DialogueEndSession(Message):
    """Sent when a dialogue session should be ended.

    You should send this message after receiving a :class:`rhasspyhermes.nlu.NluIntent` message
    if you want to end the session.

    Make sure to use the same ``sessionId`` as the original :class:`rhasspyhermes.nlu.NluIntent`
    message.

    .. admonition:: MQTT message

      Topic
        ``hermes/dialogueManager/continueSession``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - sessionId
            - String
            - The id of the session to continue.
          * - customData
            - String (optional)
            - An update to the session's custom data. If not provided, the custom data
              will stay the same.
          * - text
            - String (optional)
            - The text the TTS should say to end the session.

      Publish this message type with ``mosquitto_pub``:

      .. code-block:: shell

        mosquitto_pub -h <HOSTNAME> -t 'hermes/dialogueManager/endSession' -m '{"sessionId": "foobar", "customData": null, "text": "OK, turning off the light"}'

    Example
    -------
    >>> from rhasspyhermes.dialogue import DialogueEndSession
    >>> session = DialogueEndSession(session_id="foobar", text="OK, turning off the light")
    >>> session
    DialogueEndSession(session_id='foobar', text='OK, turning off the light', custom_data=None)
    >>> session.payload()
    '{"sessionId": "foobar", "text": "OK, turning off the light", "customData": null}'
    """

    session_id: str
    """The id of the session to end."""
    text: typing.Optional[str] = None
    """The text the TTS should say to end the session."""
    custom_data: typing.Optional[str] = None
    """An update to the session's custom data. If not provided, the custom data
    will stay the same."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"hermes/dialogueManager/endSession"``
        """
        return "hermes/dialogueManager/endSession"


@dataclass
class DialogueSessionEnded(Message):
    """Sent when a dialogue session has ended.

    .. admonition:: MQTT message

      Topic
        ``hermes/dialogueManager/sessionEnded``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - termination
            - JSON Object
            - Structured description of why the session has been ended. See
              :class:`DialogueSessionTerminationReason` for the possible values.
          * - sessionId
            - String
            - The id of the ended session.
          * - siteId
            - String
            - The id of the site where the user interaction took place.
          * - customData
            - String (optional)
            - Custom data provided in the :class:`DialogueStartSession`,
              :class:`DialogueContinueSession` or :class:`DialogueEndSession` messages.

      Subscribe to this message type with ``mosquitto_sub``:

      .. code-block:: shell

        mosquitto_sub -h <HOSTNAME> -v -t 'hermes/dialogueManager/sessionEnded'
    """

    termination: DialogueSessionTermination
    """Structured description of why the session has been ended."""
    session_id: str
    """The id of the ended session."""
    site_id: str = "default"
    """The id of the site where the user interaction took place."""
    custom_data: typing.Optional[str] = None
    """Custom data provided in the :class:`DialogueStartSession`,
    :class:`DialogueContinueSession` or :class:`DialogueEndSession` messages."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"hermes/dialogueManager/sessionEnded"``
        """
        return "hermes/dialogueManager/sessionEnded"


@dataclass
class DialogueIntentNotRecognized(Message):
    """Intent not recognized.

    Only sent when ``send_intent_not_recognized`` is ``True``.

    .. admonition:: MQTT message

      Topic
        ``hermes/dialogueManager/intentNotRecognized``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - sessionId
            - String
            - The id of the session that generated this event.
          * - siteId
            - String
            - The id of the site where the user interaction took place.
          * - input
            - String (optional)
            - The NLU input that generated this event.
          * - customData
            - String (optional)
            - Custom data provided in the :class:`DialogueStartSession` or
              :class:`DialogueContinueSession` messages.

      Subscribe to this message type with ``mosquitto_sub``:

      .. code-block:: shell

        mosquitto_sub -h <HOSTNAME> -v -t 'hermes/dialogueManager/intentNotRecognized'
    """

    session_id: str
    """The id of the session that generated this event."""
    site_id: str = "default"
    """The id of the site where the user interaction took place."""
    input: typing.Optional[str] = None
    """The NLU input that generated this event."""
    custom_data: typing.Optional[str] = None
    """Custom data provided in the :class:`DialogueStartSession` or
    :class:`DialogueContinueSession` messages."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"hermes/dialogueManager/intentNotRecognized"``
        """
        return "hermes/dialogueManager/intentNotRecognized"


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class DialogueConfigureIntent:
    """Enable/disable a specific intent in a :class:`DialogueConfigure` message."""

    intent_id: str
    """Name of the intent to enable/disable."""
    enable: bool
    """``True`` if the intent should be enabled."""


@dataclass
class DialogueConfigure(Message):
    """Enable/disable specific intents for future dialogue sessions.

    If an intent is enabled, the :class:`rhasspyhermes.nlu.NluIntent` message is triggered when
    the intent is detected.

    Rhasspy enables all intents by default unless specified otherwise.

    .. admonition:: MQTT message

      Topic
        ``hermes/dialogueManager/configure``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - intents
            - List of JSON Objects
            - The list of intents and whether to enable/disable them.
          * - siteId
            - String
            - The id of the site to configure.

      Publish this message type with ``mosquitto_pub``:

      .. code-block:: shell

        mosquitto_pub -h <HOSTNAME> -t 'hermes/dialogueManager/configure' -m '{"intents": [{"intentId": "GetTime", "enable": true}, {"intentId": "GetTemperature", "enable": false}], "siteId": "livingroom"}'

    Example
    -------
    >>> from rhasspyhermes.dialogue import DialogueConfigureIntent, DialogueConfigure
    >>> configure = DialogueConfigure([DialogueConfigureIntent("GetTime", True), DialogueConfigureIntent("GetTemperature", False)], "livingroom")
    >>> configure
    DialogueConfigure(intents=[DialogueConfigureIntent(intent_id='GetTime', enable=True), DialogueConfigureIntent(intent_id='GetTemperature', enable=False)], site_id='livingroom')
    >>> configure.payload()
    '{"intents": [{"intentId": "GetTime", "enable": true}, {"intentId": "GetTemperature", "enable": false}], "siteId": "livingroom"}'
    """

    intents: typing.List[DialogueConfigureIntent]
    """The list of intents and whether to enable/disable them."""
    site_id: str = "default"
    """The id of the site to configure."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"hermes/dialogueManager/configure"``
        """
        return "hermes/dialogueManager/configure"


# ----------------------------------------------------------------------------
# Rhasspy-Only Messages
# ----------------------------------------------------------------------------


@dataclass
class DialogueError(Message):
    """This message is published by the dialogue manager component if an error has occurred.

    .. admonition:: MQTT message

      Topic
        ``hermes/error/dialogueManager``

      Payload (JSON)
        .. list-table::
          :widths: 10 10 80
          :header-rows: 1

          * - Key
            - Type
            - Description
          * - error
            - String
            - A description of the error that occurred.
          * - siteId
            - String
            - The id of the site where the error occurred.
          * - context
            - String (optional)
            - Additional information on the context in which the error occurred.
          * - sessionId
            - String (optional)
            - The id of the session, if there is an active session.

      Subscribe to this message type with ``mosquitto_sub``:

      .. code-block:: shell

        mosquitto_sub -h <HOSTNAME> -v -t 'hermes/error/dialogueManager'

    Example
    -------

    >>> from rhasspyhermes.dialogue import DialogueError
    >>> dialogue_error = DialogueError(error="Unexpected error")
    >>> dialogue_error.topic()
    'hermes/error/dialogueManager'
    >>> dialogue_error.payload()
    '{"error": "Unexpected error", "siteId": "default", "context": null, "sessionId": null}'

    Note
    ----

    This is a Rhasspy-only message.
    """

    error: str
    """A description of the error that occurred."""
    site_id: str = "default"
    """The id of the site where the error occurred."""
    context: typing.Optional[str] = None
    """Additional information on the context in which the error occurred."""
    session_id: typing.Optional[str] = None
    """The id of the session, if there is an active session."""

    @classmethod
    def topic(cls, **kwargs) -> str:
        """Get MQTT topic for this message type.

        Returns
        -------
        str
            ``"hermes/error/dialogueManager"``
        """
        return "hermes/error/dialogueManager"
